/****************************************************************

file ger.c

Hashing package for Diophantine equations,
Version 1.0

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include<stdio.h>
#include<stdlib.h>
#include"arith.h"
#include"hashing.h"
#include"sporadic_lines.h"
#include"post_proc_kub.h"


/* Search for lines through a point. */

/* THE KNOWN "SPORADIC" LINES. */

long int  gg[GER_ANZ][3][5];


/* Solves the linear system of equations
     A0 a0 + A1 a1 + A2 a2 + A3 a3 + A4 a4 = 0
     B0 a0 + B1 a1 + B2 a2 + B3 a3 + B4 a4 = 0
   for Ai and Bi integers (of absolute value < 2**31).

   Finds three Q-linearly independent solution vectors in Z^5.

   Idea (Cramer's rule):
     a0 x + a1 y + a2 z = 0
     b0 x + b1 y + b2 z = 0
   is solved by the vector
     (     a1 a2        a0 a2       a0 a1 )
     ( det b1 b2, - det b0 b2,  det b0 b1 ).
   Replace 0 by h, 1 by i and 2 by j.

                               (ah ai)
   Find h and i such that  det (bh bi)  does not vanish.
   Then, vary j in order to generate three linearly independent
   solution vectors. */
int glei_sys (const long int A[5], const long int B[5],
                                              long int aufl[3][5]) {
 int      h, i, j, count;
 long int              d;

 /* Initialization. */
 for (i = 0; i < 3; i++)
  for (j = 0; j < 5; j++)
   aufl[i][j] = 0;

 /* Search for suitable h und i. */
 h = 0;
 while ((A[h] == 0) && (B[h] == 0)) {
  h++;
  /* The first columns are zero columns? If so then ignore them. */
  if (h >= 4) {
   /* The rank is less than 2.
      The system therefore degenerates. It is underdetermined. */
   return (-1);                         /* We give up. */
  }
 }
 i = h + 1;
 while ((A[h] * B[i] - A[i] * B[h]) == 0) {
  i++;
  if (i >= 5) {
   /* Again, the rank is less than 2.
      The system therefore degenerates. It is underdetermined. */
   return (-1);                         /* We give up. */
  }
 }
 d = A[h] * B[i] - A[i] * B[h];
 /* Now, we may resolve by xh und xi. */

 /* We put xj = d.
    xh and xi are set to be the determinants mentioned above.
    The two remaining values are set to be zero. */
 count = 0;
 for (j = 0; j <= 4; j++) {
  if ((j != h) && (j != i)) {
   aufl[count][h] = A[i]*B[j] - A[j]*B[i];
   aufl[count][i] = A[j]*B[h] - A[h]*B[j];
   aufl[count][j] = d;
   count++;
  }
 }
 /* The system is non-degenerate. */
 return 0;
}


/* The known sporadic lines are given (in sporadic_lines.h) by two
   points lying on them. This function computes another presentation
   for each of the lines consisting of three linearly independent
   equations. */
void ger_gl_init () {
 int  num;

 for (num = 0; num < GER_ANZ; num++)
  glei_sys (xyzvw[num][0], xyzvw[num][1], gg[num]);
}


/* Tests whether the the point (x : y : z : v : w) is contained in
   one of the five "standard" lines known. */
int bekannt_null (long int a, long int b,
                  long int x, long int y,
                  long int z, long int v, long int w) {
 int  num;

 for (num = 0; num < GER_ANZ; num++) {
  if ((a == ab[num][0]) && (b == ab[num][1])) {
   if ((gg[num][0][0]*x + gg[num][0][1]*y
      + gg[num][0][2]*z + gg[num][0][3]*v + gg[num][0][4]*w == 0) &&
       (gg[num][1][0]*x + gg[num][1][1]*y
      + gg[num][1][2]*z + gg[num][1][3]*v + gg[num][1][4]*w == 0) &&
      ( gg[num][2][0]*x + gg[num][2][1]*y
      + gg[num][2][2]*z + gg[num][2][3]*v + gg[num][2][4]*w == 0))
   return 1;
  }
 }
 return 0;
}


/* Tests whether the the point (x : y : z : v : w) is contained in
   one of the images under the symmetry group S_3 of the five
   "standard" lines known. */
int bekannt_eins (long int a, long int b,
                  long int x, long int y,
                  long int z, long int v, long int w) {
 if (bekannt_null (a, b, x, y, z, v, w) == 1)
  return 1;
 if (bekannt_null (a, b, x, y, z, w, v) == 1)
  return 1;
 if (bekannt_null (a, b, x, y, v, w, z) == 1)
  return 1;
 if (bekannt_null (a, b, x, y, v, z, w) == 1)
  return 1;
 if (bekannt_null (a, b, x, y, w, z, v) == 1)
  return 1;
 if (bekannt_null (a, b, x, y, w, v, z) == 1)
  return 1;
 return 0;
}


/* We consider the cubic threefold V given by
        a x**3 - b y**3 = z**3 + v**3 + w**3.
   This function returns 1 if the point (x : y : z : v : w) is
   contained in a known line on V and 0, otherwise. */
int bekannt (long int a, long int b,
             long int x, long int y,
             long int z, long int v, long int w) {
 if (a % 8 == 0)
  return (bekannt (a/8, b, 2*x, y, z, v, w));
 if (b % 8 == 0)
  return (bekannt (a, b/8, x, 2*y, z, v, w));
 if (a < b)
  return (bekannt (b, a, -y, -x, z, v, w));
 return (bekannt_eins (a, b, x, y, z, v, w));
}


/* IS THE POINT SUSPICIOUS TO LIE ON AN UNKNOWN LINE? */

/* If the contradiction is found at a prime larger than
   SCHRANKE_AUSGABE then we write an output. */
#define SCHRANKE_AUSGABE        300
/* If the contradiction is found at a prime even larger than
   SCHRANKE_VOLLE_AUSGABE then we are more verbose and output
   the point causing the contradiction, too. */
#define SCHRANKE_VOLLE_AUSGABE  300


/* The functions below are not proof against overflow when working
   with primes > 600 even on a 64-bit system. For that reason, a
   limit of 600 is hard-wired into the following tables. */
/* The primes up to 600. */
long int  primes[110], primes_inv[600];
/* Table of the square roots modulo p for the primes up to 600. */
long int             w_mod_p[110][600];
/* Table of the modular inverses for the primes up to 600. */
long int             mod_inv[600][600];
/* TO DO: There are too many global variables in this code. */


/* Tabulates the square roots modulo p. Initializes mod_inv. */
int ger_init (int lim) {
 long int  p, i, j, t;

 if (lim > 600) {
  printf ("Function ger_init called with a parameter > 600.\n");
  exit (0);
 }

 /* Initialization. */
 for (i = 0; i < 110; i++)
  primes[i] = 0;
 /* Generate the primes up to lim. */
 primes[0] = 2;
 primes[1] = 3;
 t = 3;
 for (i = 2; primes[i-1] < lim; i++) {
  do {
   t += 2;
   j = 0;
   while ((primes[j] * primes[j] <= t) && (t % primes[j] > 0))
    j++;
  } while (t % primes[j] == 0);
  primes[i] = t;
 }
 printf ("Built the table of all primes up to %ld.\n", t);

 for (j = 0; primes[j] && (primes[j] < 600); j++) {
  p = primes[j];

  primes_inv[p] = j;
  /* p is the j-th prime. */

  for (i = 0; i < p; i++)
   w_mod_p[j][i] = -1;
  for (i = 0; i <= p/2; i++)
   w_mod_p[j][(i*i) % p] = i;
  /* Table of the square roots. */

  for (i = 1; i < p; i++)
   mod_inv[i][j] = mod_inverse (i, p);
  /* Modular inverses. */
 }
 printf ("Built the table of the square roots. ");
 printf ("mod_inv initialized.\n");
 return 0;
}


/* Solves the quadratic equation A x**2 + B x + C = 0 in F_p.
   Assumes A, B, and C are represented by integers in the
   interval [0, p-1]. */
int qsolve
  (long int p, long int A, long int B, long int C, long int *loes) {
 long int  n, halb, loe, w;

 /* The equation is, in fact, at most linear. */
 if (A == 0) {
  if (B != 0) {
   loes[0] = MOD ((-C * mod_inv[B][primes_inv[p]]), p);
   /* C and mod_inv{}{} are signed integers. */
   loes[1] = -1;
   return 0;
  }

  if (C != 0) {
   loes[0] = -1;
   loes[1] = -1;
   return 0;
  }

  /* A special marker for the case A = B = C = 0 mod p. */
  loes[0] = -2;
  loes[1] = -2;
  return 0;
 }

 /* All degenerate cases have now been treated. */
 /* The p-q formula. */
 n = mod_inv[A][primes_inv[p]];
 B = (B * n) % p;
 C = (C * n) % p;

 halb = mod_inv[2][primes_inv[p]];
 loe = MOD ((-B * halb), p);
 /* B and halb are signed integers. */
 w = w_mod_p[primes_inv[p]][MOD ((loe*loe - C), p)];

 if (w == -1) {
  loes[0] = -1;
  loes[1] = -1;
  return 0;
 }

 if (w == 0) {
  loes[0] = loe;
  loes[1] = -1;
  return 0;
 }

 loes[0] = MOD ((loe + w), p);
 loes[1] = MOD ((loe - w), p);
 return 0;
}


/* Solves the linear system of equations
     A0 a0 + A1 a1 + A2 a2 + A3 a3 + A4 a4 = 0
     B0 a0 + B1 a1 + B2 a2 + B3 a3 + B4 a4 = 0
   modulo p. Finds three linearly independent solution vectors.

   Idea (Cramer's rule):
     a0 x + a1 y + a2 z = 0
     b0 x + b1 y + b2 z = 0
   is solved by the vector
     (     a1 a2        a0 a2       a0 a1 )
     ( det b1 b2, - det b0 b2,  det b0 b1 ).
   Replace 0 by h, 1 by i and 2 by j.

                               (ah ai)
   Find h and i such that  det (bh bi)  does not vanish.
   Then, vary j in order to generate three linearly independent
   solution vectors.

   ALERT: This function expects A0, ... , B4 from [0, p-1]. */
int glei_sys_mod_p (long int p, long int A[5], long int B[5],
                                           long int aufl[3][5]) {
 int      h, i, j, count;
 long int              d;

 /* Initialization. */
 for (i = 0; i < 3; i++)
  for (j = 0; j < 5; j++)
   aufl[i][j] = 0;

 /* Search for suitable h und i. */
 h = 0;
 while ((A[h] == 0) && (B[h] == 0)) {
  h++;
  /* The first columns are zero columns? If so then ignore them.
     We do not test (0 mod p) as the function is supposed to be
     called with parameters being residues modulo p. */
  if (h >= 4) {
   /* The rank over F_p is less than 2.
      The system therefore degenerates. It is underdetermined. */
   return (-1);                         /* We give up. */
  }
 }
 i = h + 1;
 while (MOD ((A[h] * B[i] - A[i] * B[h]), p) == 0) {
  i++;
  if (i >= 5) {
   /* Again, the rank over F_p is less than 2.
      The system therefore degenerates. It is underdetermined. */
   return (-1);                         /* We give up. */
  }
 }
 d = MOD ((A[h] * B[i] - A[i] * B[h]), p);
 /* Now, we may resolve by xh und xi. */

 /* We put xj = d.
    xh and xi are set to be the determinants mentioned above.
    The two remaining values are set to be zero. */
 count = 0;
 for (j = 0; j <= 4; j++) {
  if ((j != h) && (j != i)) {
   aufl[count][h] = MOD ((A[i]*B[j] - A[j]*B[i]), p);
   aufl[count][i] = MOD ((A[j]*B[h] - A[h]*B[j]), p);
   aufl[count][j] = d;
   count++;
  }
 }
 /* The system is non-degenerate. */
 return 0;
}


/* A plane is given by the three generators
         vec[0], vec[1], vec[2] \in Z^5 .
   Let further be the quadric
         q0 a0**2 + ... + q4 a4**2 = 0
   be given.

   This function computes from these data the conic which is their
   intersection. All calculations take place in F_p. */
int ebene_quadrik_mod_p (long int p, long int vec[3][5],
                               long int q[5], long int *ABCDEF) {
 long int  i, A, B, C, D, E, F;

 A = 0;
 for (i = 0; i <= 4; i++)
  A = A + q[i]*vec[0][i]*vec[0][i];
 A = MOD (A, p);

 B = 0;
 for (i = 0; i <= 4; i++)
  B = B + q[i]*vec[1][i]*vec[1][i];
 B = MOD (B, p);

 C = 0;
 for (i = 0; i <= 4; i++)
  C = C + q[i]*vec[2][i]*vec[2][i];
 C = MOD (C, p);

 D = 0;
 for (i = 0; i <= 4; i++)
  D = D + 2*q[i]*vec[0][i]*vec[1][i];
 D = MOD (D, p);

 E = 0;
 for (i = 0; i <= 4; i++)
  E = E + 2*q[i]*vec[0][i]*vec[2][i];
 E = MOD (E, p);

 F = 0;
 for (i = 0; i <= 4; i++)
  F = F + 2*q[i]*vec[1][i]*vec[2][i];
 F = MOD (F, p);

 ABCDEF[0] = A;
 ABCDEF[1] = B;
 ABCDEF[2] = C;
 ABCDEF[3] = D;
 ABCDEF[4] = E;
 ABCDEF[5] = F;

 /* The conic degenerates to the whole projective plane. */
 if ((A == 0) && (B == 0) && (C == 0) &&
     (D == 0) && (E == 0) && (F == 0))
  return 1;

 return 0;
}


/* On the conic
         A r**2 + B s**2 + C t**2 + D rs + E rt + F st = 0
   over F_p, dritte_koord finds the t-coordinates for given
   r and s. */
int dritte_koord (long int p, long int r, long int s,
                  long int A, long int B, long int C,
                  long int D, long int E, long int F, long int *t) {
 long int  a, b, c;

 a = MOD (C, p);
 b = MOD ((E*r + F*s), p);
 c = MOD ((A*r*r + B*s*s + D*r*s), p);
 qsolve (p, a, b, c, t);
 return 0;
}


/* A plane is given by the three generators
         vec[0], vec[1], vec[2] \in Z^5 .
   (r : s : t) describes a point on this plane.

   This function calculates the five coordinates back from
   r, s, and t.

   All calculations take place in F_p. */
int fuenf_koord (long int p, long int r, long int s, long int t,
                             long int vec[3][5], long int *erg) {
 int  i;

 for (i = 0; i <= 4; i++)
  erg[i] = MOD ((vec[0][i] * r + vec[1][i] * s + vec[2][i] * t), p);
 return 0;
}


/* Solves the linear system of equations modulo p by providing
   three generators for the projective plane of solutions.
   Computes the quadric in P^4 and the conic given by its
   intersection with the projective plane. */
int ebene_und_ebene_quadrik_mod_p
  (long int p, long int a, long int b, long int *x,
   long int *l0, long int *l1,
   long int (*aufl_p)[5], long int *qf_p) {
 long int  l0_p[5], l1_p[5];
 long int              q[5];
 int                      j;
 int              degen, eb;

 /* Reducing the linear system of equations modulo p. */
 for (j = 0; j <= 4; j++) {
  l0_p[j] = MOD (l0[j], p);
  l1_p[j] = MOD (l1[j], p);
 }
 /* Over F_p, the plane is given by l0_p = l1_p = 0. */

 /* Solve the linear system of equations modulo p.  */
 degen = glei_sys_mod_p (p, l0_p, l1_p, aufl_p);
 /* Modulo p, the projective plane is spanned by
    aufl_p[0], aufl_p[1], and aufl_p[2]. */

 /* The linear system of equations degenerates over F_p. */
 if (degen == -1)
  return (-1);

 /* The linear system of equations does not degenerate over F_p. */
 /* Compute the quadric in P^4. */
 q[0] = MOD ((-a * x[0]), p);
 q[1] = MOD (( b * x[1]), p);
 for (j = 2; j <= 4; j++) {
  q[j] = MOD (x[j], p);
 }
 /* Over F_p, the quadric in P^4 is given by q = 0. */

 /* Computing the conic modulo p. */
 eb = ebene_quadrik_mod_p (p, aufl_p, q, qf_p);
 /* Over F_p, the conic is given by qf_p* = 0. */

 return (eb);
}


/* Testing whether m is a possible reduction modulo p of a second
   point on the line. Actually, we only check the cubic term. */
int teste_einen_punkt (long int m[5], long int x[5],
                       long int p, long int a, long int b) {
 long int  k3;

 k3 =  -a*m[0]*m[0]*m[0]
      + b*m[1]*m[1]*m[1]
      +   m[2]*m[2]*m[2]
      +   m[3]*m[3]*m[3]
      +   m[4]*m[4]*m[4];
 k3 = MOD (k3, p);

 if (k3 != 0)
  return 0;

 /* There is no contradiction modulo p.
       (m[0] : m[1] : m[2] : m[3] : m[4])
    is (modulo p) a possible second point on the line. */
 return (-1);
}


/* Determine the linear system of equations. */
int test_init (long int a, long int b,
               long int x[5], long int l0[5], long int l1[5]) {
 int  i;

 for (i = 0; i <= 4; i++)
  l0[i] = x[i]*x[i];

 l0[0] = -a * l0[0];
 l0[1] =  b * l0[1];

 i = 4;
 while (x[i] == 0) {
  l1[i] = 0;
  i--;
  if (i == 1)
   return (-1);         /* Error at x = (x0, x1, 0, 0, 0). */
 }
 l1[i] = 1;
 i--;
 for (; i >= 0; i--)
  l1[i] = 0;

 return 0;
}


/* Testing the point x = (x0, ... , x4) modulo the prime number p.
   We assume that the plane and the conic are already computed.

   For each vector m on the conic, i.e. for each normalized vector
   which is allowed by comparison of coefficients at p**2 q
   and p q**2, a call of teste_einen_punkt occurs.

   This function expects that ger_init, test_init, and
   ebene_und_ebene_quadrik_mod_p are already run. */
int test_p (long int p, long int a, long int b, long int x[5], long int k[3][5], long int qf[6]) {
 long int  m[5];
 long int  t[2];
 long int  i, g;
 int        tmp;

 if (p == 2) {
  printf ("This function works only for p = 3 or larger.\n");
  exit (0);
 }

 /* Reminder for dritte_koord:
    t[0] >= 0 corresponds to a point which is really to be tested.
              In this case, t[1] possibly contains the third
              coordinate of a second point.
    t[0] = -1 means that we considered an unsolvable quadratic
              equation. Then, there is nothing to be tested.
    t[0] = -2 is the particular case of a trivial quadratic equation
              with all coefficients zero. */
 if ((qf[2] % p) == 0) {
 /* In this case, the point (0:0:1) lies on the conic. */
  fuenf_koord (p, 0, 0, 1, k, m);
  tmp = teste_einen_punkt (m, x, p, a, b);
  if (tmp == -1)
   return 0;
 }

 /* Points on the conic of form (0:1:x). */
 dritte_koord
          (p, 0, 1, qf[0], qf[1], qf[2], qf[3], qf[4], qf[5], t);
 /* Possible third coordinates to (0 : 1 : ?) are t[0] and t[1]. */

 /* First solution. */
 if (t[0] >= 0) {
  fuenf_koord (p, 0, 1, t[0], k, m);
  tmp = teste_einen_punkt (m, x, p, a, b);
  if (tmp == -1)
   return 0;
 }

 /* The case of a trivial equation.
    Every value of t is a solution. */
 if (t[0] <= -2) { /* Start the loop. */
  for (g = 0; g <= p-1; g++) {
   fuenf_koord (p, 0, 1, g, k, m);
   tmp = teste_einen_punkt (m, x, p, a, b);
   if (tmp == -1)
    return 0;
  }
 }

 /* Second solution. */
 if (t[1] >= 0) {
  fuenf_koord (p, 0, 1, t[1], k, m);
  tmp = teste_einen_punkt (m, x, p, a, b);
  if (tmp == -1)
   return 0;
 }

 /* Points on the conic of form (1:i:x). */
 for (i = 0; i <= p-1; i++) {
  dritte_koord
           (p, 1, i, qf[0], qf[1], qf[2], qf[3], qf[4], qf[5], t);
  /* Possible third coordinates are t[0] and t[1]. */

  /* First solution. */
  if (t[0] >= 0) {
   fuenf_koord (p, 1, i, t[0], k, m);
   tmp = teste_einen_punkt (m, x, p, a, b);
   if (tmp == -1)
    return 0;
  }

  /* The case of a trivial equation.
     Every value of t is a solution. */
  if (t[0] <= -2) { /* Start the loop. */
   for (g = 0; g <= p-1; g++) {
    fuenf_koord (p, 1, i, g, k, m);
    tmp = teste_einen_punkt (m, x, p, a, b);
    if (tmp == -1)
     return 0;
   }
  }

  /* Second solution. */
  if (t[1] >= 0) {
   fuenf_koord (p, 1, i, t[1], k, m);
   tmp = teste_einen_punkt (m, x, p, a, b);
   if (tmp == -1)
    return 0;
  }
 }

 /* The point does not lie on a line. */
 return 1;
}


/*******************************************************************
 * Testing the point (x0, x1, x2, x3, x4) subsequently modulo all
 * prime numbers up to the limit lim.
 * The function stops at the first prime p modulo which
 * unsolvability is observed. This prime is the return value.
 * If p is bigger than SCHRANKE_AUSGABE then we output p.
 * If p is even bigger than SCHRANKE_VOLLE_AUSGABE then the point
 * is output, too.
 *
 * This function expects that ger_init is already run.
 ******************************************************************/
int test (long int lim, long int a, long int b, long int *x) {
 int             i, tmp;
 long int       qf_p[6];
 long int             p;
 long int  l0[5], l1[5];
 long int  aufl_p[3][5];

 if (lim > 600) {
  printf ("Function test called with a parameter > 600.\n");
  exit (0);
 }

 /* Determine the linear system of equations over Z. */
 tmp = test_init (a, b, x, l0, l1);
 if (tmp != 0) {        /* This can't happen. */
  printf ("Error at x = (x0 : x1 : 0 : 0 : 0).\n");
  exit (0);
 }

 /* i = 2 means: Start at p = 5. */
 for (i = 2; primes[i] && ((p = primes[i]) < lim); i++) {
  tmp = ebene_und_ebene_quadrik_mod_p
                               (p, a, b, x, l0, l1, aufl_p, qf_p);
  if (tmp != 0)
   continue;
   /* Over F_p, the linear system of equations is not of full rank
      or the conic degenerates to 0 = 0.
      In the first case, we could not continue, anyway. If the conic
      degenerates to 0 = 0 then it does not make sense to test all
      points on the projective plane as we have only one cubic
      equation. */

  /* Test all the F_p-valued points on the conic.
     The hard work modulo p. */
  tmp = test_p (p, a, b, x, aufl_p, qf_p);

  if (tmp == 1) {
   /* We found a contradiction at p.
      (x[0] : x[1] : x[2] : x[3] : x[4]) certainly does not lie an
      a line.
      We will break out of the for-loop with return (p).
      Before, we possibly write an output documenting the
      contradiction. */

   /* The output. */
   if (p >= SCHRANKE_AUSGABE) {
    out (p);
    if (p >= SCHRANKE_VOLLE_AUSGABE)
     voll_out (x, a, b);
   }
   /* Contradiction. Breaking out of the for-loop. */
   return (p);
  }
 }

 /* If we end here then no contradiction was found for p up to 600.
    Therefore, the point
       (x[0] : x[1] : x[2] : x[3] : x[4])
    possibly lies on a line. */
 return 0;
}


/* Output functions. */
/* Necessary only to make the output look more nicely. */
int  cursor_pos;


/* Cursor to the left. */
void cursor_init () {
 cursor_pos = 0;
}


/* Output a suspicious point. */
void verd_out
  (long int x, long int y, long int z, long int v, long int w,
   long int a, long int b) {
 if (cursor_pos == 0)
  printf ("\n");
 printf ("Suspicious point ");
 printf ("(%ld : %ld : %ld : %ld : %ld), ", x, y, z, v, w);
 printf ("a = %ld, b = %ld.\n", a, b);
 cursor_pos = 0;
}


/* Output a (large) prime at which a contradiction was found. */
void out (long int p) {
 printf ("%ld ", p);
 cursor_pos = 1;
}


/* Output a(n even larger) prime at which a contradiction was found.
   Output the point considered, too. */
void voll_out (long int *x, long int a, long int b) {
 printf ("\nPoint ");
 printf ("(%ld : %ld : %ld : %ld : %ld), ",
                                   x[0], x[1], x[2], x[3], x[4]);
 printf ("a = %ld, b = %ld.\n", a, b);
 cursor_pos = 0;
}
