/****************************************************************

file kub.c

Hashing package for Diophantine equations,
Version 1.0

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include<stdio.h>
#include<stdlib.h>
#include"arith.h"
#include"hashing.h"
#include"post_proc_kub.h"


/* Table of the third power residues modulo PAGE_PRIME
   from 0**3 through (SUCHWEITE-1)**3. */
unsigned int                                             *h3_mod_p;
unsigned int                           h3_mod_p_roh[2*SUCHWEITE+1];

/* Table of the third roots modulo PAGE_PRIME. */
signed int                              dritte_wurzeln[PAGE_PRIME];

/* a**{-1} modulo PAGE_PRIME for a in [1, A_LIMIT]. */
long int                                            inv[A_LIMIT+1];


/* init does the following.
   A pre-test:     If PAGE_PRIME is unsuitable => Early exit.
   Tabulate:       Multiplicative inverse, third powers and
                   third roots modulo PAGE_PRIME. */
void init () {
 long int  i, j;

 /* Testing PAGE_PRIME. */
 if (PAGE_PRIME > SUCHWEITE) {
  printf ("PAGE_PRIME must not be bigger than SUCHWEITE.\n");
  exit (0);
 }
 if (PAGE_PRIME <= A_LIMIT) {
  printf ("PAGE_PRIME must be bigger than A_LIMIT.\n");
  exit (0);
 }
 if ((PAGE_PRIME % 3) != 2) {
  printf ("PAGE_PRIME must be 2 modulo 3.\n");
  exit (0);
 }

 /* Computation of the inverses modulo PAGE_PRIME. */
 for (i = 1; i <= A_LIMIT; i++) {
  inv[i] = mod_inverse (i, PAGE_PRIME);
 }

 /* Adjusting the pointer.
    h3_mod_p allows indices from -SUCHWEITE to SUCHWEITE. */
 h3_mod_p = &h3_mod_p_roh[SUCHWEITE];

 /* Tabulating the third powers modulo PAGE_PRIME. */
 for (j = -SUCHWEITE; j <= SUCHWEITE; j++)
  h3_mod_p[j] = MOD ((j*j*j), PAGE_PRIME);

 /* Tabulating the third roots modulo PAGE_PRIME.
    This code assumes PAGE_PRIME <= SUCHWEITE. */
 for (i = 0; i < PAGE_PRIME; i++)
  dritte_wurzeln[i] = -1;
 for (j = 0; j < PAGE_PRIME; j++)
  dritte_wurzeln[h3_mod_p[j]] = j;
}


/* WRITING */

/* Writes the triple (v, w, z) into the hash table. */
inline void z_v_und_w_schreiben
    (long int z, long int v, long int w) {
 long int                     rechte_seite;
 struct nebentabellen_eintrag          dat;

 /* There should be no overflow over \pm 2**63. */
 rechte_seite = z*z*z + v*v*v + w*w*w;

 dat.z = z;
 dat.v = v;
 dat.w = w;

 element_schreiben (dat, rechte_seite);
}


/* The outer loop of the writing part for residue class rkl. */
void schreibdaten_erzeugen_und_schreiben (unsigned long int rkl) {
 long int  v, w, z, z0, z03;

 for (w = -SUCHWEITE + 1; w < SUCHWEITE; w++) {
  for (v = w; v < SUCHWEITE; v++) {             /* Only v >= w. */
   z03 = (rkl + 2*PAGE_PRIME - h3_mod_p[v] - h3_mod_p[w])
                                                     % PAGE_PRIME;
   /* Residue class of z**3 modulo PAGE_PRIME. */
   z0 = dritte_wurzeln[z03];

   if (z0 != -1) {
   /* If there exists a third root of z03. */
    z = z0;
    while (z > v)
     z -= PAGE_PRIME;
    while (z < v)
     z += PAGE_PRIME;                           /* Only z >= v. */
    while (z < SUCHWEITE) {
     z_v_und_w_schreiben (z, v, w);
     z += PAGE_PRIME;
    }
   } /* End-if to if (z0 != -1)                */
  }  /* End to for (v = w; v < SUCHWEITE; v++) */
 }   /* End to for (w = 0; w < SUCHWEITE; w++) */
}


/* READING */

/* Look up (a, b, x, y) in the hash table. */
inline void a_b_x_und_y_testen (long int a, long int b,
                                long int x, long int y) {
 long int                rechte_seite;
 struct variablen_rechts          akt;

 akt.x = x; akt.y = y;
 akt.a = a; akt.b = b;

 rechte_seite = a*x*x*x - b*y*y*y;
 /* There should be no overflow over \pm 2**63. */

 /* Calls post_proc whenever necessary. post_proc includes the
    search for lines. */
 element_lesen (akt, rechte_seite, post_proc);
}


/* The outer loop of the reading part for residue class rkl. */
void lesedaten_erzeugen_und_lesen (unsigned long int rkl) {
 long int  a, b, x, y, x0, x03;

 for (a = 1; a <= A_LIMIT; a++) {
  for (b = 1; (b <= a) && (b <= B_LIMIT); b++) {
   for (y = -SUCHWEITE + 1; y < SUCHWEITE; y++) {
    x03 = (rkl + b * h3_mod_p[y]) % PAGE_PRIME;
    /* Residue class of a * x**3 modulo PAGE_PRIME. */
    x03 = (x03 * inv[a]) % PAGE_PRIME;
    /* Residue class of x**3 modulo PAGE_PRIME. */
    x0 = dritte_wurzeln[x03];

    if (x0 != -1) { /* If there exists a third root of x03. */
     x = x0;
     while (x < 0)
      x += PAGE_PRIME;      /* Only solutions such that x >= 0. */
     while (x < SUCHWEITE) {
      a_b_x_und_y_testen (a, b, x, y);
      x += PAGE_PRIME;
     }
    } /* End-if to if (x0 != -1).                             */
   }  /* End to for (y = -SUCHWEITE + 1; y < SUCHWEITE; y++). */
  }   /* End to for (b = 1; (b <= a) && (b <= B_LIMIT); b++). */
 }    /* End to for (a = 1; a <= A_LIMIT; a++).               */
}


/* The outer loop. */
int main (int argc, char *argv[]) {
 unsigned long int  rkl, start, ende;

 if (argc != 3) {
  printf ("Call needs 2 parameters.\n");
  exit (0);
 }

 start = atol (argv[1]);
 ende = atol (argv[2]);
 if (start < 0)
  start = 0;
 if (ende > (PAGE_PRIME - 1))
  ende = PAGE_PRIME - 1;
 if (start > ende) {
  printf ("Range from %ld to %ld is empty.\n", start, ende);
  exit (0);
 }

 printf ("Range from %ld bis %ld.\n", start, ende);
 printf ("Initialization of the tables of powers and roots.\n");
 /* Initialize the cursor position. */
 cursor_init ();
 init ();
 hashing_init ();
 /* Initialize the arrays for the numbers of solutions. */
 counter_init ();
 /* Initialization of the machinery searching for lines. */
 ger_init (600);
 /* This function does not stand a parameter > 600. */
 ger_gl_init ();
 printf ("Initialization finished.\n");

 for (rkl = start; rkl <= ende; rkl++) {
  printf ("\nBuilding up the hash table in residue class %ld.\n",
   rkl);
  schreiben_init ();
  schreibdaten_erzeugen_und_schreiben (rkl);
  schreiben_beenden ();
  printf ("Searching in residue class %ld.\n", rkl);

  lesen_init ();
  lesedaten_erzeugen_und_lesen (rkl);
  /* Calls post_proc whenever necessary. post_proc includes the
     search for lines. */
  lesen_beenden (post_proc);

  fflush (stdout);
 }
 hashing_beenden ();
 counter_out ();

 return 0;
}
