/****************************************************************

file post_proc_kub.c

Hashing package for Diophantine equations,
Version 1.0

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include<stdio.h>
#include<stdlib.h>
#include"arith.h"
#include"hashing.h"
#include"post_proc_kub.h"


/* POST PROCESSING */

#define INTERVALL_ANZ (1 + SUCHWEITE / INTERVALL_LAENGE)

/* Number of coincidences in the hash table.
   Number of solutions found for a pair (a, b) of height within a
   subinterval of [1, SUCHWEITE] of length INTERVALL_LAENGE. */
long int              anz_aus[A_LIMIT+1][B_LIMIT+1][INTERVALL_ANZ];

/* Extra counters for points on obvious lines, for points on known
   lines, and for points which are suspicious. The latter means,
   they probably lie on a non-obvious, unknown line. */
long int
       anz_auf_offens_geraden[A_LIMIT+1][B_LIMIT+1][INTERVALL_ANZ];
long int  anz_auf_gef_geraden[A_LIMIT+1][B_LIMIT+1][INTERVALL_ANZ];
long int      anz_verdaechtig[A_LIMIT+1][B_LIMIT+1][INTERVALL_ANZ];


/* Initialization of the arrays for the numbers of solutions. */
void counter_init () {
 unsigned long int  a, b, i;

 for (a = 0; a <= A_LIMIT; a++) {
  for (b = 0; b <= B_LIMIT; b++) {
   for (i = 0; i < INTERVALL_ANZ; i++) {
    anz_aus[a][b][i] = 0;
    anz_auf_offens_geraden[a][b][i] = 0;
    anz_auf_gef_geraden[a][b][i] = 0;
    anz_verdaechtig[a][b][i] = 0;
   }
  }
 }
}


/* Output the final values of all counters. */
void counter_out () {
 unsigned long int  a, b, i;

 for (a = 1; a <= A_LIMIT; a++) {
  for (b = 1; (b <= a) && (b <= B_LIMIT); b++) {
   printf ("a = %ld, b = %ld: ", a, b);
   for (i = 0; i < INTERVALL_ANZ; i++) {
    printf (" %ld ", anz_aus[a][b][i]/2);
   }
   printf (" solutions, among them ");
   for (i = 0; i < INTERVALL_ANZ; i++) {
    printf (" %ld ", anz_auf_offens_geraden[a][b][i]/2);
   }
   printf (" on obvious lines and");
   for (i = 0; i < INTERVALL_ANZ; i++) {
    printf (" %ld ", anz_auf_gef_geraden[a][b][i]/2);
   }
   printf (" on known lines.");
   printf (" Further, ");
   for (i = 0; i < INTERVALL_ANZ; i++) {
    printf (" %ld ", anz_verdaechtig[a][b][i]/2);
   }
   printf (" suspicious points found.\n");
  }
 }
}


/* Test whether the point (x : y : z : v : w) lies on one of the
   lines. If so, increase the corresponding counter. */
void geradentest (long int a, long int b,
                  long int x, long int y,
                  long int z, long int v, long int w,
                  int bereich, int mult) {
 int          tmp;
 long int  vec[5];

 /* The obvious lines -- ten real elliptic cones. */
 if (( a*x*x*x == b*y*y*y) || ( a*x*x*x == z*z*z) ||
     ( a*x*x*x == v*v*v)   || ( a*x*x*x == w*w*w) ||
     (-b*y*y*y == z*z*z)   || (-b*y*y*y == v*v*v) ||
     (-b*y*y*y == w*w*w)   ||        (z == -v)    ||
            (z == -w)      ||        (v == -w)) {
  anz_auf_offens_geraden[a][b][bereich] += mult;
 }
 else {
  /* Testing whether the point lies on a line although it is outside
     the elliptic cones. */
  if (bekannt (a, b, x, y, z, v, w)) /* The other known lines. */
   anz_auf_gef_geraden[a][b][bereich] += mult;
  else {
   /* Is the point suspicious to lie on an unknown line? */
   vec[0] = x;
   vec[1] = y;
   vec[2] = z;
   vec[3] = v;
   vec[4] = w;
   /* This function does not stand a first parameter > 600. */
   tmp = test (600, a, b, vec);
   if (tmp == 0) {
    verd_out (x, y, z, v, w, a, b);
    anz_verdaechtig[a][b][bereich] += mult;
   }
  }
 }
}


/* Post processing. Did we really detect a solution?
   If so, write an output, increase the counter, and call
   geradentest. */
void post_proc
  (struct nebentabellen_eintrag nt, struct variablen_rechts r) {
 long int        a, b, x, y, z, v, w;
 long int  i, j, fak, hoehe, bereich;

 a = r.a; b = r.b; x = r.x; y = r.y;
 z = nt.z;  v = nt.v;  w = nt.w;

 /* There should be no overflow over \pm 2**63. */
 if (a*x*x*x - b*y*y*y == z*z*z + v*v*v + w*w*w) {
  /* Test whether the solution found is primitive. */
  fak = gcd (x, y);
  if (fak != 1) {
   fak = gcd (fak, z);
   if (fak != 1) {
    fak = gcd (fak, v);
    if (fak != 1)
     fak = gcd (fak, w);
   }
  }

  if (fak == 1) {
   hoehe = MAX (MAX (MAX (labs (x), labs (y)),
                              MAX (labs (z), labs (v))), labs (w));
   bereich = hoehe / INTERVALL_LAENGE;
   /* Solution (x, y, z, v, w) found for (a, b). */

   if (z > v) {
    if (v > w) i = 6; else i = 3;
   }
   else {
    if (v > w) i = 3; else i = 1;
   } /* Now, i is the number of the permutations of (z, v, w). */

   j = 2;
   if (x == 0)
    j = 1;
   /* i * j is twice the number of solutions in projective space.
      If (x, y, z, v, w) is a solution then (-x, -y, -z, -v, -w)
      is another one.
      Both together should count as one.
      However, if x = 0 then we find them both while for x \neq 0
      we find only one of them.
      Consequently, solutions such that x = 0 have to be counted
      with a weight of one half. */

   anz_aus[a][b][bereich] += (i * j);
   geradentest (a, b, x, y, z, v, w, bereich, i*j);
  }
 }
}
