/****************************************************************

file hashing.c

Hashing package for Diophantine equations,
Version 1.0

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include<stdio.h>
#include"hashing.h"

#define HASH_GRENZE ((1U << HASH_BREITE) - 1)


/*  The hash function. */
#define HASH(a)        (((unsigned int) (a)) >> (32 - HASH_BREITE))

/* CONTROL_FILTER = 2**n - 1 where n <= 15.
   For n = 16, the control value could become negative and -1 would
   lose the extra meaning "Place unoccupied". */
#define CONTROL_FILTER ((1U << 15) - 1)
/* The control function. */
#define CONTROL(a)                                                \
                ((signed short int) (((a) >> 32) & CONTROL_FILTER))


/* Data structures for the prefetch buffers. */
struct lesepuffer_eintrag {
 struct variablen_rechts    r;
 int                     h, c;
};

struct schreibpuffer_eintrag {
 struct nebentabellen_eintrag   nt;
 int                          h, c;
};

/* The hash table. */
signed short int                        hash_c[HASH_GRENZE];
struct nebentabellen_eintrag      nebentabelle[HASH_GRENZE];

/* The prefetch buffer for reading. */
struct lesepuffer_eintrag               lese_pu[PUFFER_LEN];
/* A pointer to the prefetch buffer. */
int                                                     lpz;

/* The prefetch buffer for writing. */
struct schreibpuffer_eintrag         schreib_pu[PUFFER_LEN];
/* A pointer to the prefetch buffer. */
int                                                     spz;

/* Number of entries written into the hash table. */
unsigned long int                                    h_used;

/* Total number of reading attempts in the hash table. */
unsigned long int                                    h_read;

/* Total number of coincidences. */
unsigned long int                                 anz_coinc;


/* Initialize two counters. */
void hashing_init () {
 h_read = 0;
 anz_coinc = 0;
}


/* Output some statistics. */
void hashing_beenden () {
 printf ("%ld attempts to read from the hash table.\n", h_read);
 printf ("Totally %ld coincidences.\n", anz_coinc);
}


/* WRITING */

/* Initialization --
   Declare the hash table and the input buffer to be empty. */
void schreiben_init () {
 unsigned long int  i;

 h_used = 0;
 for (i = 0; i <= HASH_GRENZE; i++)
  hash_c[i] = -1;

 spz = 0;
 for (i = 0; i < PUFFER_LEN; i++)
  schreib_pu[i].c = -2;
  schreib_pu[i].h = 0;
  /* -2 is an unused symbol. It is our marker for an empty place. */
}


/* Writes into the hash table via an input buffer.
   We write the control value into hash_c.
   The struct nebentabellen_eintrag containing information about the
   variables on the left hand side is written into nebentabelle. */
inline void element_schreiben
   (struct nebentabellen_eintrag dat,
    unsigned long int linke_seite_l) {
 unsigned int  t;

 if (schreib_pu[spz].c != -2) { /* Not an empty place. */
  t = schreib_pu[spz].h;

  /* Resolution of collisions. */
  while (hash_c[t] != -1)
   t = ((t + 1) & HASH_GRENZE);    /* (t+1) mod (2**HASH_BREITE) */

  hash_c[t] = schreib_pu[spz].c;
  nebentabelle[t] = schreib_pu[spz].nt;
 }

 /* Write the nebentabellen_eintrag dat into the buffer at the free
    position spz. */
 schreib_pu[spz].c = CONTROL (linke_seite_l);
 t = HASH (linke_seite_l); schreib_pu[spz].h = t;
 #if (WITH_PREFETCH)
  __asm__ ("prefetchnta (%0)" : : "r" ((long int) &hash_c[t]));
  __asm__ ("prefetchnta (%0)" : : "r" ((long int) &hash_c[t + 2]));
 #endif
 schreib_pu[spz].nt = dat;

 /* Now point to the next position in the buffer. */
 /* spz = (spz + 1) % PUFFER_LEN; */
 spz++;
 if (spz >= PUFFER_LEN) spz = 0;

 h_used++;
}


/* Write the final entries of the input buffer into the hash table.
   Then, declare the buffer to be empty.
   Output some statistics. */
void schreiben_beenden () {
 int j, t;

 for (j = 0; j < PUFFER_LEN; j++) {
  if (schreib_pu[j].c != -2) { /* Not an empty place. */
   t = schreib_pu[j].h;

   /* Resolution of collisions. */
   while (hash_c[t] != -1)
   t = ((t + 1) & HASH_GRENZE);    /* (t+1) mod (2**HASH_BREITE) */

   hash_c[t] = schreib_pu[j].c;
   nebentabelle[t] = schreib_pu[j].nt;

   schreib_pu[j].c = -2; /* Mark as empty. */
  }
 }

 printf ("%ld entries in the hash table ", h_used);
 printf ("(Size %d).\n", HASH_GRENZE);
}


/* READING */

/* Initialization --
   Declare the reading buffer to be empty. */
void lesen_init () {
 unsigned long int  i;

 lpz = 0;
 for (i = 0; i < PUFFER_LEN; i++) {
  lese_pu[i].h = 0;
  lese_pu[i].c = -2;
  /* -2 is an unused symbol. There will certainly be no
     coincidences. */
 }
}


/* Writes the present variablen_rechts into the reading buffer.
   Calls prefetchnta.
   Takes an old variablen_rechts from the reading buffer and
   searches in the hash table for a coincidence. */
inline void element_lesen
  (struct variablen_rechts dat, unsigned long int rechte_seite,
   pp_fun_pointer pp_fun) {
 signed int    c;
 unsigned int  t;

 h_read++;

 /* Reading from the buffer at position lpz. */
 t = lese_pu[lpz].h;
 c = lese_pu[lpz].c;

 /* Optimization for speed. Some sort of loop unrolling.
    If we hit an unoccupied place then that reading access finishes
    directly at this line. */
 if (hash_c[t] != -1) {
  do { /* Resolution of collisions */
   /* We found variablen_rechts containing interesting values. */
   if (hash_c[t] == c) {
    anz_coinc++;
    pp_fun (nebentabelle[t], lese_pu[lpz].r);
   }
   t = ((t + 1) & HASH_GRENZE);    /* (t+1) mod (2**HASH_BREITE) */
  }
  while (hash_c[t] != -1);
 } /* End-if to if hash_c[t] >= 0. */

 /* Write the variablen_rechts dat into the buffer at the free
    position lpz. */
 lese_pu[lpz].c = CONTROL (rechte_seite);
 t = HASH (rechte_seite); lese_pu[lpz].h = t;
 #if (WITH_PREFETCH)
  __asm__ ("prefetchnta (%0)" : : "r" ((long int) &hash_c[t]));
  __asm__ ("prefetchnta (%0)" : : "r" ((long int) &hash_c[t + 2]));
 #endif
 lese_pu[lpz].r = dat;

 /* Now point to the next position in the buffer. */
 /* lpz = (lpz + 1) % PUFFER_LEN; */
 lpz++;
 if (lpz >= PUFFER_LEN) lpz = 0;
}


/* Finishes all open reading accesses.

   At the end, the buffer is certainly full (possibly with -2's).
   Therefore, we need an empty pass. Here, we write nonsense into
   the buffer. But, possibly, useful things are found. */
void lesen_beenden (pp_fun_pointer pp_fun) {
 unsigned long int          i;
 struct variablen_rechts  dat;

 for (i = 0; i < PUFFER_LEN; i++)
  element_lesen (dat, 0, pp_fun);
}
