/****************************************************************

file post_proc_quart.c

Hashing package for Diophantine equations,
Version 1.0

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include<stdio.h>
#include"arith.h"
#include"hashing.h"
#include"post_proc_quart.h"


/* POST PROCESSING */

/* Number of coincidences in the hash table.
   Number of solutions found for the pair (a, b). */
long int  anz_aus[A_LIMIT+1][B_LIMIT+1];


/* Initialization of the array for the numbers of solutions. */
void counter_init () {
 unsigned long int  a, b;

 for (a = 0; a <= A_LIMIT; a++)
  for (b = 0; b <= B_LIMIT; b++)
   anz_aus[a][b] = 0;
}


/* Output the numbers of solutions. */
void counter_out () {
 unsigned long int  a, b;

 for (a = 1; a <= A_LIMIT; a++)
  for (b = 1; b <= B_LIMIT; b++)
   printf ("a = %ld, b = %ld: %ld solutions.\n",
                                              a, b, anz_aus[a][b]);
}


/* Computes the length of the orbit of a point in projective space
   und the operation of the 96-element symmetry group of the
   quartic. */
unsigned long int bahnlaenge
   (unsigned long int x, unsigned long int y,
    unsigned long int z, unsigned long int v, unsigned long int w) {
 unsigned long int  i, j;

 if (z != v) {
  if (v != w) i = 6; else i = 3;
 }
 else {
  if (v != w) i = 3; else i = 1;
 } /* Now, i is the number of the permutations of (z, v, w). */

 j = 1;
 if (x != 0) j <<= 1;
 if (y != 0) j <<= 1;
 if (z != 0) j <<= 1;
 if (v != 0) j <<= 1;
 if (w != 0) j <<= 1;
 j >>= 1;

 /* Number of the generated solutions with signs allowed. */
 return (i * j);
}


/* Reconstruction of x and y from the sparing table.

   0xfL is the same as 15 as a long int. Therefore,
        (mix_xy_high[t] & 0xfL) << 16
   are the final 4 binary digits of mix_xy_high[t]
   multiplied with 2**16.

   Correspondingly,
        ((unsigned long int) mix_xy_high[t] >> 4) << 16
   are the next to last 4 binary digits of mix_xy_high[t]
   multiplied with 2**16. */
inline struct alle_variablen
alle_variablen_rekonstruieren
  (struct nebentabellen_eintrag tab, struct variablen_rechts puf) {
 struct alle_variablen  liste;

 liste.l.a = tab.a;
 liste.l.b = tab.b;
 liste.l.x = (((unsigned long int) tab.x)
                | ((tab.mix_xy_high & 0xfL) << 16));
 liste.l.y = (((unsigned long int) tab.y)
                | (((unsigned long int)
                    tab.mix_xy_high >> 4) << 16));
 liste.r = puf;
 return (liste);
}


/* post_proc is rather complicated but not called very often. */
void post_proc
  (struct nebentabellen_eintrag tab, struct variablen_rechts puf) {
 unsigned long int     a, b, x, y, z, v, w;
 unsigned long int                     fak;
 struct alle_variablen               liste;

 liste = alle_variablen_rekonstruieren (tab,  puf);
 a = liste.l.a;
 b = liste.l.b;
 /* liste.l.a and liste.l.b are unsigned chars.
    This code should work for 0 < a, b < 256. */
 x = liste.l.x;
 y = liste.l.y;
 z = liste.r.z;
 v = liste.r.v;
 w = liste.r.w;

 /* Test modulo 2**64.
    This suffices as we also have the congruence modulo PAGE_PRIME.
    Indeed, by our estimates for the size, the left hand side is
    in the interval
       [0, 3 * SUCHWEITE**4] .
    The right hand side belongs to that interval anyway.
    Note that 3 * SUCHWEITE**4 < 2**64 * PAGE_PRIME, by
    assumption. */
 if (a*x*x*x*x - b*y*y*y*y == z*z*z*z + v*v*v*v + w*w*w*w) {
  /* Test whether the solution found is primitive. */
  fak = gcd (x, y);
  if (fak != 1) {
   fak = gcd (fak, z);
   if (fak != 1) {
    fak = gcd (fak, v);
    if (fak != 1) {
     fak = gcd (fak, w);
    }
   }
  }

  if (fak == 1) {
   #if (WANT_OUTPUT_EACH_SOLUTION)
    printf ("(a, b) = (%ld, %ld), ", a, b);
    printf ("Height %ld, ", MAX (MAX (MAX (x, y), MAX (z, v)), w));
    printf ("Solution (%ld, %ld, %ld, %ld, %ld).\n", x, y, z, v, w);
   #endif

   /* Counting with the correct multiplicity. */
   anz_aus[a][b] += bahnlaenge (x, y, z, v, w);
  } /* End-if to if (fak == 1). */
 }  /* End-if to if (a*x*x*x*x - b*y*y*y*y; ... ). */
}
