/****************************************************************

file quart.c

Hashing package for Diophantine equations,
Version 1.0

Copyright (C) 2006 A.-S. Elsenhans and J. Jahnel

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Address of the authors

	A.-S. Elsenhans/ J. Jahnel
	Math. Institut der Universitaet
	Bunsenstrasze 3--5
	D-37073 Goettingen
        Germany

WWW     http://www.uni-math.gwdg.de/jahnel

*****************************************************************/


#include<stdio.h>
#include<stdlib.h>
#include"arith.h"
#include"hashing.h"
#include"post_proc_quart.h"


/* The value 3 * SUCHWEITE**4 as a 128-bit integer represented by
   two 64-bit integers. */
unsigned long int                              max_h, max_l;

/* Table of the fourth powers modulo PAGE_PRIME
   from 0**4 through (SUCHWEITE-1)**4. */
unsigned short int                      h4_mod_p[SUCHWEITE];

/* Table of the fourth roots modulo PAGE_PRIME. */
unsigned short int               vierte_wurzeln[PAGE_PRIME];

/* a**{-1} modulo PAGE_PRIME for a in [1, A_LIMIT]. */
long int                                     inv[A_LIMIT+1];


/* init does the following.
   A pre-test:     If PAGE_PRIME is unsuitable => Early exit.
   Tabulate:       Multiplicative Inverse, fourth powers and fourth
                   roots modulo PAGE_PRIME.
   Initialize:     max_h and max_l as the high and low limb,
                   respectively, of the 128 bit integer
                   3 * SUCHWEITE**4. */
void init () {
 unsigned long int  a, i, j;
 unsigned long int        t;

 /* The biggest value represented by z**4 + v**4 + w**4. */
 umul_ppmm (max_h, max_l,
 3L * (unsigned long int) SUCHWEITE * (unsigned long int) SUCHWEITE,
     (unsigned long int) SUCHWEITE * (unsigned long int) SUCHWEITE);

 /* Testing PAGE_PRIME. */
 if (PAGE_PRIME <= A_LIMIT) {
  printf ("PAGE_PRIME must be bigger than A_LIMIT.\n");
  exit (0);
 }
 if (PAGE_PRIME > 65535) {
  printf ("PAGE_PRIME must not be bigger than 65536.\n");
  exit (0);
 }
 if ((PAGE_PRIME % 4) != 3) {
  printf ("PAGE_PRIME must be 3 modulo 4.\n");
  exit (0);
 }
 if (max_h >= PAGE_PRIME) {
  printf ("2**64*PAGE_PRIME must be bigger than 3*SUCHWEITE**4.\n");
  exit (0);
 }
 for (i = 3; i*i <= PAGE_PRIME; i += 2) {
  if (PAGE_PRIME % i == 0) {
   printf ("PAGE_PRIME must a prime number.\n");
   exit (0);
  }
 }

 /* Computation of the inverses modulo PAGE_PRIME. */
 for (a = 1; a <= A_LIMIT; a++) {
  inv[a] = mod_inverse (a, PAGE_PRIME);
 }

 /* Tabulate the fourth powers modulo PAGE_PRIME. */
 for (j = 0; j < SUCHWEITE; j++) {
  t = (j * j) % PAGE_PRIME;
  h4_mod_p[j] = (t * t) % PAGE_PRIME;
 }

 /* Tabulate the fourth roots modulo PAGE_PRIME. */
 for (i = 0; i < PAGE_PRIME; i++)
  vierte_wurzeln[i] = ~0;
 /* ~0 = 2**16 - 1. Our special symbol for an unoccupied place. */
 for (j = 0; (j < SUCHWEITE) && (j < PAGE_PRIME); j++)
  vierte_wurzeln[h4_mod_p[j]] = j;
 /* If both fourth roots are bigger than SUCHWEITE then no fourth
    root will be found. */
}


/* WRITING */

/* Writes the quadruple (a, b, x, y) into the hash table. */
inline void a_b_x_und_y_schreiben
  (unsigned long int a, unsigned long int b,
   unsigned long int x, unsigned long int y,
   unsigned long int linke_seite_l) {
 struct nebentabellen_eintrag  dat;

 dat.a = a;
 dat.b = b;
 dat.x = (unsigned short int) x;
 dat.y = (unsigned short int) y;
       /* The 16 least significant bits of x and y, respectively. */
 dat.mix_xy_high = (x >> 16) | ((y >> 16) << 4);
       /* And here, there are bits number 17 through 20. */
 element_schreiben (dat, linke_seite_l);
}


/* Compares the size exactly
   (a rough comparison of the size being assumed)
   and tests the congruence modulo 5 or 625, respectively.
   Writes into the hash table if the tests were successful. */
inline void test_und_schreiben
  (unsigned long int a, unsigned long int b,
   unsigned long int x, unsigned long int y,
   unsigned long int linke_seite_h,
   unsigned long int linke_seite_l) {
 int  m625, m5;

 /* Exact comparison under the assumption of rough comparison. */
 if ((linke_seite_h < max_h) || (linke_seite_l <= max_l)) {
  /* Test of linke_seite modulo 625. Write only if
     (linke_seite mod 625) is represented by z**4 + v**4 + w**4. */
  m625 = (linke_seite_h * 366 + (linke_seite_l % 625L)) % 625;
  /* 366 = 2**64 mod 625. */
  m5 = m625 % 5;
  if (m5 == 0) {
   if ((m625 == 0) && ((x % 5 != 0) || (y % 5 != 0)))
    a_b_x_und_y_schreiben (a, b, x, y, linke_seite_l);
  }
  else {
   if (m5 != 4)
    a_b_x_und_y_schreiben (a, b, x, y, linke_seite_l);
  }
 }
}


/* Writes the quadruple (a, b, x, y) into the hash table if
   a x**4 - b y**4
    - is positive,
    - not by far too big,
    - suitable modulo 5 (tested in function test_und_schreiben),
    - and leads, in the case a*x*x*x*x - b*y*y*y*y = 0, to a
      primitive solution. */
inline void quadrupel_a_b_x_y_behandeln
      (unsigned long int a, unsigned long int b,
       unsigned long int x, unsigned long int y) {
 unsigned long int  linke_seite_h, linke_seite_l;
 unsigned long int   ggt, ax4h, ax4l, by4h, by4l;

 /* linke_seite = a*x*x*x*x - b*y*y*y*y modulo 2**128. */
 umul_ppmm (ax4h, ax4l, a*x*x, x*x);
 umul_ppmm (by4h, by4l, b*y*y, y*y);
 sub_ddmmss (linke_seite_h, linke_seite_l, ax4h, ax4l, by4h, by4l);

 /* If 0 <= linke_seite <= 3 * SUCHWEITE**4 (rough comparison). */
 if ((linke_seite_h <= max_h)) {
  /* Special treatment of the case linke_seite = 0.
     Necessary in order to avoid bad clustering. */
  if ((linke_seite_l == 0) && (linke_seite_h == 0)) {
   ggt = gcd (x, y);
   if (ggt == 1)
    a_b_x_und_y_schreiben (a, b, x, y, linke_seite_l);
  }
  else
   test_und_schreiben (a, b, x, y, linke_seite_h, linke_seite_l);
 }
}


/* The outer loop of the writing part for residue class rkl. */
void schreibdaten_erzeugen_und_schreiben (unsigned long int rkl) {
 unsigned long int  a, b, x, y, x04, y0;
 long int                            x0;
 int                       xg, xu, diff;

 for (a = 1; a <= A_LIMIT; a++) {
  for (b = 1; b <= B_LIMIT; b++) {
   /* We deal with the cases of even and odd y, separately. */
   for (y0 = 0; y0 < 2; y0++) {
    /* Does it make sense to consider even values of x? Odd ones? */
    if        ((a + 15*b*y0) % 16 <= 3 ) xu = 1; else xu = 0;
    if ((y0 == 1) && ((15*b) % 16 <= 3)) xg = 1; else xg = 0;
    /* If only one of the parities makes sense then take longer
       steps. */
    if ((xu == 0) || (xg == 0))
     diff = 2 * PAGE_PRIME;
    else
     diff = PAGE_PRIME;

    if (xg | xu) {
     for (y = y0; y < SUCHWEITE; y += 2) {
      x04 = (rkl + b * h4_mod_p[y]) % PAGE_PRIME;
      /* Residue class of a * x**4 modulo PAGE_PRIME. */
      x04 = (x04 * inv[a]) % PAGE_PRIME;
      /* Residue class of x**4 modulo PAGE_PRIME. */
      x0 = vierte_wurzeln[x04];

      if (x0 != (unsigned short int) ~0) {
       /* If there exists a fourth root of x04. */
       x = x0;
       if ((x % 2 == 0) && (xg == 0)) x += PAGE_PRIME;
       if ((x % 2 == 1) && (xu == 0)) x += PAGE_PRIME;
       while (x < SUCHWEITE) {
        quadrupel_a_b_x_y_behandeln (a, b, x, y);
        x += diff;
       }
       if (x0 != 0) {
       /* If x0 = 0 then there exists only one fourth root. */
        x = PAGE_PRIME - x0;      /* The other fourth root. */
        if ((x % 2 == 0) && (xg == 0)) x += PAGE_PRIME;
        if ((x % 2 == 1) && (xu == 0)) x += PAGE_PRIME;
        while (x < SUCHWEITE) {
         quadrupel_a_b_x_y_behandeln (a, b, x, y);
         x += diff;
        } /* End to while (x < SUCHWEITE).              */
       }  /* End-if to if (x0 != 0).                    */
      }   /* End-if to if (x0 != -1).                   */
     }    /* End to for (y = 0; y < SUCHWEITE; y += 2). */
    }     /* End-if to if (xg + xu > 0).                */
   }      /* End to for (y0 = 0; y0 < 2; y0++).         */
  }       /* End to for (b = 1; b <= B_LIMIT; b++).     */
 }        /* End to for (a = 1; a <= A_LIMIT; a++).     */
}


/* READING */

/* Look up (z, v, w) in the hash table. */
inline void z_v_und_w_testen
  (unsigned long int z, unsigned long int v, unsigned long int w) {
 struct variablen_rechts          dat;
 unsigned long int       rechte_seite;

 dat.z = z;
 dat.v = v;
 dat.w = w;

 rechte_seite = z*z*z*z + v*v*v*v + w*w*w*w; /* modulo 2**64. */
 element_lesen (dat, rechte_seite, post_proc);
}


/* Test the triples (z, v, w) where z runs through the residue class
   (z0 mod PAGE_PRIME). */
inline void z0_v_und_w_testen
  (unsigned long int z0, unsigned long int v, unsigned long int w) {
 unsigned long int  z;

 #if (SUCHWEITE < 2 * PAGE_PRIME)
  /* Loop-unrolling -- Done by hand. */
  z = z0;
  if (z <= v) {         /* Only z <= v. */
   z_v_und_w_testen (z, v, w);
   z += PAGE_PRIME;
   if (z <= v) {
    z_v_und_w_testen (z, v, w);
   }
  }
 #else
  while (z <= v) {      /* Only z <= v. */
   z_v_und_w_testen (z, v, w);
   z += PAGE_PRIME;
  }
 #endif
}


/* The outer loop of the reading part for residue class rkl. */
void lesedaten_erzeugen_und_lesen (unsigned long int rkl) {
 unsigned long int  v, w, z04;
 unsigned short int        z0;

 for (w = 0; w < SUCHWEITE; w++) {
  for (v = 0; v <= w; v++) {            /* Only v <= w. */
   z04 = (rkl + 2*PAGE_PRIME - h4_mod_p[w] - h4_mod_p[v]);
   /* % PAGE_PRIME; */
   if (z04 >= 2 * PAGE_PRIME) z04 -= 2 * PAGE_PRIME;
   if (z04 >= PAGE_PRIME) z04 -= PAGE_PRIME;
   /* The residue class of z**4 modulo PAGE_PRIME. */

   z0 = vierte_wurzeln[z04];

   /* If there exists a fourth root of z04. */
   if (z0 != (unsigned short int) ~0) {
    z0_v_und_w_testen (z0, v, w);

    /* If z0 = 0 then there is only one fourth root.
       Otherwise, work with the other fourth root of z04, too. */
    if (z0 != 0)
     z0_v_und_w_testen (PAGE_PRIME - z0, v, w);
   } /* End-if to if (z0 != -1).                */
  }  /* End to for (v = w; v <= w; v++).        */
 }   /* End to for (w = 0; w < SUCHWEITE; w++). */
}


/* The outer loop. */
int main (int argc, char *argv[]) {
 unsigned long int  rkl, start, ende;

 if (argc != 3) {
  printf ("Call needs 2 parameters.\n");
  exit (0);
 }
 start = atol (argv[1]);
 ende = atol (argv[2]);
 if (start < 0)
  start = 0;
 if (ende > (PAGE_PRIME - 1))
  ende = PAGE_PRIME - 1;
 if (start > ende) {
  printf ("Range from %ld to %ld is empty.\n", start, ende);
  exit (0);
 }

 printf ("Range from %ld to %ld.\n", start, ende);

 hashing_init ();
 /* Initialize the arrays for the numbers of solutions. */
 counter_init ();
 printf ("Initialization of the tables of powers and roots.\n");
 init ();
 printf ("Initialization finished.\n");

 for (rkl = start; rkl <= ende; rkl++) {
  printf ("Building up the hash table in residue class %ld.\n",
   rkl);
  schreiben_init ();
  schreibdaten_erzeugen_und_schreiben (rkl);
  schreiben_beenden ();

  printf ("Searching in residue class %ld.\n", rkl);
  lesen_init ();
  lesedaten_erzeugen_und_lesen (rkl);
  lesen_beenden (post_proc);

  fflush (stdout);
 }
 counter_out ();
 hashing_beenden ();

 return 0;
}
